/*!
 * \file  io-format.cc
 */

#include "config.h"

#include "pnapi-assert.h"

#include "automaton.h"
#include "interface.h"
#include "petrinet.h"
#include "state.h"
#include "myio.h"
#include "util.h"

#include <sstream>

using std::cerr;
using std::endl;
using std::map;
using std::ostream;
using std::set;
using std::string;
using std::stringstream;

using pnapi::io::util::delim;
using pnapi::io::util::filterMarkedPlaces;
using pnapi::io::util::Mode;
using pnapi::io::util::ModeData;

using namespace pnapi::formula;

namespace pnapi
{

namespace io
{
using pnapi::io::util::operator<<;


/* FORMAT IMPLEMENTATION: add a corresponding section */

/*************************************************************************
 ***** STAT output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ostream & stat(std::ostream & ios)
{
  util::FormatData::data(ios) = util::STAT;
  return ios;
}


namespace __stat
{

/*!
 * \brief output petri net statistics
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  set<Label *> asynchronous = net.interface_.getAsynchronousLabels();
  
  unsigned int placeCount = net.places_.size() + asynchronous.size();
  unsigned int arcCount = net.arcs_.size();
  
  PNAPI_FOREACH(l, asynchronous)
  {
    arcCount += (*l)->getTransitions().size();
  }
  
  return (os << "|P|= " << placeCount << "  "
             << "|P_in|= " << net.interface_.getInputLabels().size() << "  "
             << "|P_out|= " << net.interface_.getOutputLabels().size() << "  "
             << "|T|= " << net.transitions_.size() << "  "
             << "|F|= " << arcCount);
}

/*!
 * \brief write service automaton statistics
 */
std::ostream & output(std::ostream & os, const Automaton & sa)
{
  return os << "|Q|= " << sa.states_.size() << "  "
            << "|E|= " << sa.edges_.size();
}

} /* namespace __stat */


/*************************************************************************
 ***** DOT output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ostream & dot(std::ostream & ios)
{
  util::FormatData::data(ios) = util::DOT;
  return ios;
}


namespace __dot
{

/*!
 * \brief output net 
 * 
 * Creates a DOT output (see http://www.graphviz.org) of the net.
 * It uses the digraph-statement and adds labels to transitions,
 * places and arcs if neccessary.
 * 
 * \todo print final condition below net
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  string filename = net.getMetaInformation(os, io::INPUTFILE);
  
  // initialize node name cache
  unsigned int dotID = 0;
  set<Place *> places = net.getPlaces();
  PNAPI_FOREACH(it, places)
  {
    stringstream ss; ss << "p" << (++dotID);
    util::DotNameData::data(os).names[(*it)->getName()] = ss.str();
  }

  dotID = 0;
  set<Transition *> transitions = net.getTransitions();
  PNAPI_FOREACH(it, transitions)
  {
    stringstream ss; ss << "t" << ++dotID;
    util::DotNameData::data(os).names[(*it)->getName()] = ss.str();
  }
  
  dotID = 0;
  set<Label *> input = net.getInterface().getInputLabels();
  PNAPI_FOREACH(it, input)
  {
    stringstream ss; ss << "i" << (++dotID);
    util::DotNameData::data(os).names[(*it)->getName()] = ss.str();
  }

  dotID = 0;
  set<Label *> output = net.getInterface().getOutputLabels();
  PNAPI_FOREACH(it, output)
  {
    stringstream ss; ss << "o" << (++dotID);
    util::DotNameData::data(os).names[(*it)->getName()] = ss.str();
  }
  
  dotID = 0;
  set<Label *> synchronous = net.getInterface().getSynchronousLabels();
  PNAPI_FOREACH(it, synchronous)
  {
    stringstream ss; ss << "s" << (++dotID);
    util::DotNameData::data(os).names[(*it)->getName()] = ss.str();
  }


  os  //< output everything to this stream

  << delim("\n")
  << util::mode(io::util::NORMAL)

  << "digraph N {\n"
  << " graph [fontname=\"Helvetica\" nodesep=0.25 ranksep=\"0.25\""
  << " remincross=true label=\""
  << "Petri net"
  << (filename.empty() ? "" : " generated from " + filename) << "\\n"
  << "&Omega;=" << net.getFinalCondition() << "\"]\n"
  
  << " node [fontname=\"Helvetica\" fixedsize width=\".3\""
  << " height=\".3\" label=\"\" style=filled]\n"
  << " edge [fontname=\"Helvetica\" color=white arrowhead=none"
  << " weight=\"20.0\"]\n\n"
  
  << net.interface_

  << "// places\n"
  << " node [shape=circle fillcolor=white]\n"
  << net.places_

  << "\n // transitions\n"
  << " node [shape=box]\n"
  << net.transitions_

  << util::mode(io::util::INNER)
  << delim(" ")
  << "\n\n // inner cluster\n"
  << " subgraph cluster1\n"
  << " {\n"
  << "  " << net.transitions_ << endl
  << "  " << net.places_ << endl
  << "  label=\"\" style=" << (net.interface_.isEmpty() ? "invis" : "\"dashed\"")
  << "\n }\n\n"

  << delim("\n")
  << util::mode(io::util::ARC)
  << " // arcs\n"
  << " edge [fontname=\"Helvetica\" arrowhead=normal color=black]\n"
  << net.arcs_ << endl
  << util::mode(io::util::PLACE)
  // actually not an appropriate mode but I did not want to introduce 
  // a further mode like e.g. "ARC2" for only this purpose.   
  << net.transitions_ // interface arcs

  << "\n}\n";
  
  return os;
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  // place attributes (used in NORMAL mode)
  stringstream attributes;
  if (p.getTokenCount() == 1)
  {
    if(p.getColor() == "")
    {
      attributes << "fillcolor=black";
    }
    else
    {
      attributes << "fillcolor=" << p.getColor();
    }
    attributes << " peripheries=2 height=\".2\" "
               << "width=\".2\" ";
  }
  else
  {
    if(p.getColor() != "")
    {
      attributes << "fillcolor=" << p.getColor() << " ";
    }
    
    if (p.getTokenCount() > 1)
    {
      attributes << "label=\"" << p.getTokenCount() << "\" "
                 << "fontcolor=black fontname=\"Helvetica\" ";
    }
  }

  if (p.wasInterface())
  {
    attributes << "fillcolor=lightgoldenrod1";
  }

  // output the place as a node
  return output(os, p, attributes.str());
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  if(ModeData::data(os) == util::PLACE) // mode when printing the arcs to synchronous labels; actually should be something like "ARC2"
  {
    PNAPI_FOREACH(l, t.getLabels())
    {
      string left;
      string right;
      stringstream attr;
      attr << "\t[weight=10000.0";
      
      if(l->second != 1)
      {
        attr << " label=\"" << (l->second) << "\"";
      }
      
      switch(l->first->getType())
      {
      case Label::SYNCHRONOUS:
        attr << " arrowhead=none";
      case Label::INPUT:
        left = getLabelName(os, *l->first);
        right = getNodeName(os, t);
        attr << "]\n";
        break;
      case Label::OUTPUT:
        left = getNodeName(os, t);
        right = getLabelName(os, *l->first);
        attr << "]\n";
        break;
      default: PNAPI_ASSERT(false);
      }
      
      os << " " << left << " -> " << right << attr.str();
    }
    
    return os;
  }
  
  // transition attributes (used in NORMAL mode)
  stringstream attributes;
  if(ModeData::data(os) != util::ARC) // when printing arcs, no attributes are needed
  {
    switch(t.getType())
    {
    case(Transition::INPUT):  attributes << "fillcolor=orange"; break;
    case(Transition::OUTPUT): attributes << "fillcolor=yellow"; break;
    case(Transition::INOUT):
    {
      attributes
      << "fillcolor=gold label=< <TABLE BORDER=\"1\""
      << " CELLBORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"0\""
      << " HEIGHT=\"21\" WIDTH=\"21\" FIXEDSIZE=\"TRUE\"><TR>"
      << "<TD HEIGHT=\"11\" WIDTH=\"21\" FIXEDSIZE=\"TRUE\""
      << " BGCOLOR=\"ORANGE\">"
      << "</TD></TR><TR>"
      << "<TD HEIGHT=\"10\" WIDTH=\"21\" FIXEDSIZE=\"TRUE\""
      << " BGCOLOR=\"YELLOW\">"
      << "</TD></TR></TABLE> >";
      
      break;
    }
    default:
      if (t.getColor() != "")
      {
        attributes << "fillcolor=" << t.getColor();
      }
    }
  }

  // output the transition as a node
  output(os, t, attributes.str());

  return os;
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  os << " " << arc.getSourceNode() << " -> " << arc.getTargetNode() 
     << "\t[";
  
  if (arc.getWeight() != 1)
  {
    os << "label=\"" << arc.getWeight() << "\"";
  }
  
  return (os << "]");
}

/*!
 * \brief interface output
 */
std::ostream & output(std::ostream & os, const Interface & interface)
{
  os << "// interface\n" << delim("\n");
  
  map<string, Port *> ports = interface.getPorts();
  
  if(util::PortData::data(os).remove)
  {
    os << "/// input\n node [shape=circle fillcolor=orange]\n"<< interface.getInputLabels()
       << "\n/// output\n node [shape=circle fillcolor=yellow]\n" << interface.getOutputLabels()
       << "\n/// synchronous\n node [shape=box fillcolor=black]\n" << interface.getSynchronousLabels()
       << endl;
  }
  else
  {
    PNAPI_FOREACH(port, ports)
    {
      os << (*port->second) << endl;
    }
  }
  
  return (os << endl);
}

/*!
 * \brief port output
 */
std::ostream & output(std::ostream & os, const Port & port)
{
  static PetriNet * net = NULL;
  static int portNumber = 0;
  
  if(net != &port.getPetriNet())
  {
    net = &port.getPetriNet();
    portNumber = 0;
  }
  
  util::Mode oldMode = ModeData::data(os);
  
  os << mode(util::NORMAL) << "/// port '" << port.getName() 
     << "'\n//// input\n node [shape=circle fillcolor=orange]\n" << port.getInputLabels()
     << "\n//// output\n node [shape=circle fillcolor=yellow]\n" << port.getOutputLabels()
     << "\n//// synchronous\n node [shape=box fillcolor=black]\n" << port.getSynchronousLabels() 
     << "\n//// cluster\n subgraph cluster_port" << (++portNumber) << "\n {\n  "
     << mode(util::INNER) << delim(" ")
     << port.getInputLabels() << "\n  "
     << port.getOutputLabels() << "\n  "
     << port.getSynchronousLabels()
     << "\n  label=\"" << port.getName() << "\" style=\"filled,rounded\"; bgcolor=grey95  labelloc=t;\n }\n\n"
     << mode(oldMode) << delim("\n");
     
  return os;
}

/*!
 * \brief label output
 */
std::ostream & output(std::ostream & os, const Label & label)
{
  switch(ModeData::data(os))
  {
  case util::NORMAL:
  {
    string labelName =  getLabelName(os, label);
    string labelName_l = getLabelName(os, label, true);
    
    os << labelName
       << ((label.getType() == Label::SYNCHRONOUS) ? "\t[width=.1]\n  " : "\n  ")
       << labelName_l << "\t[style=invis]\n  "
       << labelName << " -> " << labelName_l
       << "\t[headlabel=\"" << label.getName() << "\"]";
    
    break;
  }
  case util::INNER:
  {
    os << getLabelName(os, label) << " " << getLabelName(os, label, true);
    break;
  }
    
  default: /* do nothing */ ;
  }
  return os;
}

/*!
 * \brief output node with attribute
 */
std::ostream & output(std::ostream & os, const Node & n, const std::string & attr)
{
  string dotName   = getNodeName(os, n);
  string dotName_l = getNodeName(os, n, true);

  os << dotName;

  switch(ModeData::data(os))
  {
  case util::NORMAL:
  {
    os << "\t[" << attr << "]" << endl
       << " " << dotName_l << "\t[style=invis]" << endl
       << " " << dotName_l << " -> " << dotName
       << " [headlabel=\"" << n.getName() << "\"]";
    
    break;
  }
  case util::INNER:
  {
    os << " " << dotName_l;
    break;
  }
  default: /* do nothing */ ;
  }

  return os;
}


/*!
 * \brief get a unique node name for dot output
 * 
 * \pre all nodes and labels have distinct names
 */
std::string getNodeName(std::ostream & os, const Node & n, bool withSuffix)
{
  return util::DotNameData::data(os).names[n.getName()] + string(withSuffix ? "_l" : "");
}

/*!
 * \brief get a unique label name for dot output
 * 
 * \pre all nodes and labels have distinct names
 */
std::string getLabelName(std::ostream & os, const Label & l, bool withSuffix)
{
  return util::DotNameData::data(os).names[l.getName()] + string(withSuffix ? "_l" : "");
}

/*!
 * \brief service automaton output
 */
std::ostream & output(std::ostream & os, const Automaton & sa)
{
  os << "Digraph ServiceAutomaton {\n"
     << "{\n"
     << "q0 [style=invis];\n";
  
  for(int i = 0; i < (int) sa.states_.size(); ++i)
  {
    if (sa.states_[i]->isFinal())
      os << sa.states_[i]->getName() << " [];\n";
    else
      os << sa.states_[i]->getName() << ";\n";
  }
  
  os << "}\n"
     << "{\n"
     << "q0 -> " << (*sa.getInitialStates().begin())->getName() << ";\n";
  
  for(int i = 0; i < (int) sa.edges_.size(); ++i)
  {
    os << sa.edges_[i]->getSource().getName() << " -> "
       << sa.edges_[i]->getDestination().getName() << " [label=\""
       << sa.edges_[i]->getLabel() <<"\"];\n";
  }
  
  return (os << "}\n}\n");
}

/*!
 * \brief negation output
 */
std::ostream & output(std::ostream & os, const formula::Negation & f)
{
  return (os << "&not;" << (*f.getChildren().begin())); 
}

/*!
 * \brief conjunction output
 */
std::ostream & output(std::ostream & os, const formula::Conjunction & f)
{
  return (os << "(" << delim(" &and; ") << f.getChildren() << ")");
}

/*!
 * \brief disjunction output
 */
std::ostream & output(std::ostream & os, const formula::Disjunction & f)
{
  return (os << "(" << delim(" &or; ") << f.getChildren() << ")");
}

/*!
 * \brief FormulaTrue output
 */
std::ostream & output(std::ostream & os, const formula::FormulaTrue &)
{
  return (os << "True");
}

/*!
 * \brief FormulaFalse output
 */
std::ostream & output(std::ostream & os, const formula::FormulaFalse &)
{
  return (os << "False");
}

/*!
 * \brief FormulaEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaEqual & f)
{
  return (os << "(" << f.getPlace().getName() << "=" << f.getTokens() << ")");
}

/*!
 * \brief FormulaNotEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaNotEqual & f)
{
  return (os << "(" << f.getPlace().getName() << "&ne;" << f.getTokens() << ")");
}

/*!
 * \brief FormulaLess output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLess & f)
{
  return (os << "(" << f.getPlace().getName() << "&lt;" << f.getTokens() << ")");
}

/*!
 * \brief FormulaLessEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLessEqual & f)
{
  return (os << "(" << f.getPlace().getName() << "&le;" << f.getTokens() << ")");
}

/*!
 * \brief FormulaGreater output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreater & f)
{
  return (os << "(" << f.getPlace().getName() << "&gt;" << f.getTokens() << ")");
}

/*!
 * \brief FormulaGreaterEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreaterEqual & f)
{
  return (os << "(" << f.getPlace().getName() << "&ge;" << f.getTokens() << ")");
}


} /* namespace __dot */



/*************************************************************************
 ***** LOLA output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & lola(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::LOLA;
  return ios;
}

/*!
 * \brief makes the stream write also final condition formulae
 */
std::ostream & formula(std::ostream & os)
{
  util::FormulaData::data(os).formula = true;
  return os;
}


namespace __lola
{

/*!
 * \brief petri net output
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  string creator = net.getMetaInformation(os, CREATOR, PACKAGE_STRING);
  string inputfile = net.getMetaInformation(os, INPUTFILE);

  os //< output everything to this stream

  << "{ Petri net created by " << creator
  << (inputfile.empty() ? "" : " reading " + inputfile)
  << " }\n\n"

  << "PLACE\n  " << mode(io::util::PLACE) << delim(", ")
  << net.places_ << ";\n\n"

  << "MARKING\n  " << mode(io::util::PLACE_TOKEN)
  << filterMarkedPlaces(net.places_) << ";\n\n\n"

  // transitions
  << delim("\n") << net.transitions_ << endl
  << endl;

  if (util::FormulaData::data(os).formula)
  {
    os << "FORMULA\n  "
       << net.finalCondition_ << endl
       << endl;
  }

  return (os << "{ END OF FILE }\n");
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  os << p.getName();
  
  if(ModeData::data(os) == io::util::PLACE_TOKEN)
  {
    os << ":" << p.getTokenCount();
  }
  
  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  
  return (os << "TRANSITION " << t.getName()
            << delim(", ")
            << "\n  CONSUME "
            << t.getPresetArcs() << ";\n"
            << "  PRODUCE "
            << t.getPostsetArcs() << ";\n");
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  return (os << arc.getPlace().getName() << ":" << arc.getWeight());
}

/*!
 * \brief negation output
 */
std::ostream & output(std::ostream & os, const formula::Negation & f)
{
  if (f.getChildren().empty())
    throw exception::NotImplementedError("don't know how to print a negation of nothing");
  else
    return (os << "NOT (" << **f.getChildren().begin() << ")");
}

/*
 * \brief conjunction output
 */
std::ostream & output(std::ostream & os, const formula::Conjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaTrue();
    throw exception::NotImplementedError("don't know how to print an empty conjunction");
  }
  else
    return (os << "(" << delim(" AND ") << f.getChildren() << ")");
}


/*!
 * \brief disjunction output
 */
std::ostream & output(std::ostream & os, const formula::Disjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaFalse();
    throw exception::NotImplementedError("don't know how to print an empty disjunction");
  }
  else
    return (os << "(" << delim(" OR ") << f.getChildren() << ")");
}

/*!
 * \brief output FormulaTrue
 */
std::ostream & output(std::ostream & os, const formula::FormulaTrue &)
{
  return (os << "TRUE");  // keyword not yet implemented in lola
  // TODO: remove comment when done
}

/*!
 * \brief output FormulaFalse
 */
std::ostream & output(std::ostream & os, const formula::FormulaFalse &)
{
  return (os << "FALSE"); // keyword not yet implemented in lola
  // TODO: remove comment when done
}

/*!
 * \brief output FormulaEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaEqual & f)
{
  return (os << f.getPlace().getName() << " = " << f.getTokens());
}

/*!
 * \brief output FormulaNotEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaNotEqual & f)
{
  return (os << f.getPlace().getName() << " # " << f.getTokens());
}

/*!
 * \brief output FormulaGreater
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreater & f)
{
  return (os << f.getPlace().getName() << " > " << f.getTokens());
}

/*!
 * \brief output FormulaGreaterEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreaterEqual & f)
{
  return (os << f.getPlace().getName() << " >= " << f.getTokens());
}

/*!
 * \brief output FormulaLess
 */
std::ostream & output(std::ostream & os, const formula::FormulaLess & f)
{
  return (os << f.getPlace().getName() << " < " << f.getTokens());
}

/*!
 * \brief output FormulaLessEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaLessEqual & f)
{
  return (os << f.getPlace().getName() << " <= " << f.getTokens());
}

} /* namespace __lola */


/*************************************************************************
 ***** IFN output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & ifn(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::IFN;
  return ios;
}


namespace __ifn
{

/*!
 * \brief petri net output
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  string creator = net.getMetaInformation(os, CREATOR, PACKAGE_STRING);
  string inputfile = net.getMetaInformation(os, INPUTFILE);

  os //< output everything to this stream

  << "{ Petri net created by " << creator
  << (inputfile.empty() ? "" : " reading " + inputfile)
  << " }\n\n"

  << "PLACE\n  " << mode(io::util::PLACE) << delim(", ")
  << net.places_ << ";\n\n"

  << "MARKING\n  " << mode(io::util::PLACE_TOKEN)
  << filterMarkedPlaces(net.places_) << ";\n\n\n"

  // transitions
  << delim("\n") << net.transitions_ << endl
  << endl;

  if (util::FormulaData::data(os).formula)
  {
    os << "FORMULA\n  "
       << net.finalCondition_ << endl
       << endl;
  }

  return (os << "{ END OF FILE }\n");
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  os << p.getName();
  
  if(ModeData::data(os) == io::util::PLACE_TOKEN)
  {
    os << ":" << p.getTokenCount();
  }
  
  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  string confidence = "";
  
  if (t.getConfidence() == 1) {confidence = " LOW";}
  if (t.getConfidence() == 2) {confidence = " HIGH";}
  if (t.getConfidence() == 3) {confidence = " DOWN";}
  
  return (os << "TRANSITION " << t.getName()
            << confidence
            << delim(", ")
            << "\n  CONSUME "
            << t.getPresetArcs() << ";\n"
            << "  PRODUCE "
            << t.getPostsetArcs() << ";\n");
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  return (os << arc.getPlace().getName() << ":" << arc.getWeight());
}

/*!
 * \brief negation output
 */
std::ostream & output(std::ostream & os, const formula::Negation & f)
{
  if (f.getChildren().empty())
    throw exception::NotImplementedError("don't know how to print a negation of nothing");
  else
    return (os << "NOT (" << **f.getChildren().begin() << ")");
}

/*
 * \brief conjunction output
 */
std::ostream & output(std::ostream & os, const formula::Conjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaTrue();
    throw exception::NotImplementedError("don't know how to print an empty conjunction");
  }
  else
    return (os << "(" << delim(" AND ") << f.getChildren() << ")");
}


/*!
 * \brief disjunction output
 */
std::ostream & output(std::ostream & os, const formula::Disjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaFalse();
    throw exception::NotImplementedError("don't know how to print an empty disjunction");
  }
  else
    return (os << "(" << delim(" OR ") << f.getChildren() << ")");
}

/*!
 * \brief output FormulaTrue
 */
std::ostream & output(std::ostream & os, const formula::FormulaTrue &)
{
  return (os << "TRUE");  // keyword not yet implemented in lola
  // TODO: remove comment when done
}

/*!
 * \brief output FormulaFalse
 */
std::ostream & output(std::ostream & os, const formula::FormulaFalse &)
{
  return (os << "FALSE"); // keyword not yet implemented in lola
  // TODO: remove comment when done
}

/*!
 * \brief output FormulaEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaEqual & f)
{
  return (os << f.getPlace().getName() << " = " << f.getTokens());
}

/*!
 * \brief output FormulaNotEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaNotEqual & f)
{
  return (os << f.getPlace().getName() << " # " << f.getTokens());
}

/*!
 * \brief output FormulaGreater
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreater & f)
{
  return (os << f.getPlace().getName() << " > " << f.getTokens());
}

/*!
 * \brief output FormulaGreaterEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreaterEqual & f)
{
  return (os << f.getPlace().getName() << " >= " << f.getTokens());
}

/*!
 * \brief output FormulaLess
 */
std::ostream & output(std::ostream & os, const formula::FormulaLess & f)
{
  return (os << f.getPlace().getName() << " < " << f.getTokens());
}

/*!
 * \brief output FormulaLessEqual
 */
std::ostream & output(std::ostream & os, const formula::FormulaLessEqual & f)
{
  return (os << f.getPlace().getName() << " <= " << f.getTokens());
}

} /* namespace __ifn */


/*************************************************************************
 ***** PNML output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & pnml(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::PNML;
  return ios;
}


namespace __pnml
{

/*!
 * \brief petri net output
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  os //< output everything to this stream

  << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n\n"
  << "<!--\n"
  << "  generator:   " << net.getMetaInformation(os, CREATOR, PACKAGE_STRING) << endl
  << "  input file:  " << net.getMetaInformation(os, INPUTFILE) << endl
  << "  invocation:  " << net.getMetaInformation(os, INVOCATION) << endl
  << "  net size:    " << stat << net << pnml  
  << "\n-->\n\n"

  << "<pnml>\n"
  
  << "  <module>\n"

  << net.interface_

  << "\n    <net id=\"n1\" type=\"PTNet\">\n";

  if (!net.getMetaInformation(os, INPUTFILE).empty())
  {
    os << "    <name>\n"
       << "      <text>" << net.getMetaInformation(os, INPUTFILE) << "</text>\n"
       << "    </name>\n";
  }

  os << net.places_

  << net.transitions_

  << net.arcs_

  << "    </net>\n"

  << "    <finalmarkings>\n"
  << "      <marking>\n"
  << net.finalCondition_
  << "      </marking>\n"  
  << "    </finalmarkings>\n"

  << "  </module>\n"
  
  << "</pnml>\n";

  return os << endl;
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  os << "      <place id=\"" << p.getName() << "\"";
  
  if (p.getTokenCount())
  {
    os << ">\n"
       << "        <initialMarking>\n"
       << "          <text>" << p.getTokenCount() << "</text>\n"
       << "        </initialMarking>\n"
       << "      </place>\n";
  } else {
    os << " />\n";
  }

  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  os << "      <transition id=\"" << t.getName();
  
  if(t.getLabels().empty())
  {
    return (os << "\" />\n");
  }
  
  os << "\">\n";
  
  PNAPI_FOREACH(label, t.getLabels())
  {
    os << "        <";
    switch(label->first->getType())
    {
    case Label::INPUT: os << "receive"; break;
    case Label::OUTPUT: os << "send"; break;
    case Label::SYNCHRONOUS: os << "synchronize"; break;
    default: PNAPI_ASSERT(false);
    }
    
    os << " idref=\"" + label->first->getName() + "\" />\n";
  }
  
  return (os << "      </transition>\n");
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  static PetriNet * net = NULL;
  static unsigned int arcId = 0;
  
  // check for different nets
  if(net != &arc.getPetriNet())
  {
    net = &arc.getPetriNet();
    arcId = 0;
  }
  
  os << "      <arc id=\"arcId" << (++arcId)
     << "\" source=\"" << arc.getSourceNode().getName()
     << "\" target=\"" << arc.getTargetNode().getName()
     << "\"";

  if (arc.getWeight() > 1)
  {
    os << ">\n"
       << "        <inscription>\n"
       << "          <text>" << arc.getWeight() << "</text>\n"
       << "        </inscription>\n"
       << "      </arc>\n";
  }
  else
  {
    os << " />\n";
  }

  return os;
}

/*!
 * \brief interface output
 */
std::ostream & output(std::ostream & os, const Interface & interface)
{
  os << "    <ports>\n";
  
  PNAPI_FOREACH(port, interface.getPorts())
  {
    os << "      <port id=\"" << port->first << "\">\n"
       << (*port->second)
       << "      </port>\n";
  }
    
  return (os << "    </ports>\n");
}

/*!
 * \brief port output
 */
std::ostream & output(std::ostream & os, const Port & port)
{
  return (os << delim("\n") << port.getAllLabels() << delim("") << "\n");
}

/*!
 * \brief label output
 */
std::ostream & output(std::ostream & os, const Label & l)
{
  os << "        <";
  
  switch(l.getType())
  {
  case Label::INPUT: os << "input"; break;
  case Label::OUTPUT: os << "output"; break;
  case Label::SYNCHRONOUS: os << "synchronous"; break;
  default: PNAPI_ASSERT(false); 
  }
  
  return (os << " id=\"" << l.getName() << "\" />");
}

/*!
 * \brief negation output
 */
std::ostream & output(std::ostream & os, const formula::Negation & f)
{
  if(f.getChildren().empty())
    throw exception::NotImplementedError("don't know how to print a negation of nothing");
  else
    return (os << "NOT (" << (**f.getChildren().begin()) << ")");
}

/*!
 * \brief Conjunction output
 */
std::ostream & output(std::ostream & os, const formula::Conjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaTrue();
    throw exception::NotImplementedError("don't know how to print an empty conjunction");
  }
  else
    return (os << f.getChildren());
}

/*!
 * \brief Disjunction output
 */
std::ostream & output(std::ostream & os, const formula::Disjunction & f)
{
  if(f.getChildren().empty())
  {
    //return os << formula::FormulaFalse();
    throw exception::NotImplementedError("don't know how to print an empty disjunction");
  }
  else
    return (os << delim("      </marking>\n      <marking>\n") << f.getChildren());
}

/*!
 * \brief FormulaTrue output
 */
std::ostream & output(std::ostream & os, const formula::FormulaTrue &)
{
    return os;// << "TRUE";  // TODO: nothing to write?
}

/*!
 * \brief FormulaFalse output
 */
std::ostream & output(std::ostream & os, const formula::FormulaFalse &)
{
  return os;// << "FALSE"; // TODO: nothing to write?
}

/*!
 * \brief FormulaEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaEqual & f)
{
  return (os << "        <place idref=\"" << f.getPlace().getName()
             << "\">\n          <text>" << f.getTokens() 
             << "</text>\n       </place>\n");
}

/*!
 * \brief FormulaNotEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaNotEqual & f)
{
  return (os << f.getPlace().getName() << " # " << f.getTokens());
}

/*!
 * \brief FormulaGreater output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreater & f)
{
  return (os << f.getPlace().getName() << " > " << f.getTokens());
}

/*!
 * \brief FormulaGreaterEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreaterEqual & f)
{
  return (os << f.getPlace().getName() << " >= " << f.getTokens());
}

/*!
 * \brief FormulaLess output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLess & f)
{
  return (os << f.getPlace().getName() << " < " << f.getTokens());
}

/*!
 * \brief FormulaLessEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLessEqual & f)
{
  return (os << f.getPlace().getName() << " <= " << f.getTokens());
}

} /* namespace __pnml */



/*************************************************************************
 ***** OWFN output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & owfn(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::OWFN;
  return ios;
}

/*!
 * \brief suppress role output
 */
std::ostream & noRoles(std::ostream & os)
{
  util::RoleData::data(os).role = true;
  return os;
}

/*!
 * \brief remove all ports
 */
std::ostream & removePorts(std::ostream & os)
{
  util::PortData::data(os).remove = true;
  return os;
}


namespace __owfn
{

/*!
 * \brief petri net output 
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  os  //< output everything to this stream

  << "{\n  generated by: "
  << net.getMetaInformation(os, io::CREATOR, PACKAGE_STRING)
  << "\n  input file:   "
  << net.getMetaInformation(os, io::INPUTFILE)
  << "\n  invocation:   "
  << net.getMetaInformation(os, io::INVOCATION)
  << "\n  net size:     "
  << stat << net << owfn
  << "\n}\n\n"

  << net.interface_
  
  << util::mode(io::util::PLACE) << delim(";\n  ")
  << "PLACE\n  "
  << io::util::groupPlacesByCapacity(net.places_)
  << ";\n\n" << delim(", ");
  
  if(!net.roles_.empty() && !(util::RoleData::data(os).role))
  {
    os << "ROLES\n  " << net.roles_ << ";\n\n";  
  }
  
  os
  << mode(io::util::PLACE_TOKEN)
  << "INITIALMARKING\n  "
  << io::util::filterMarkedPlaces(net.places_) << ";\n\n"

  << "FINALCONDITION\n  ";
  
  // if formel is total
  if(net.finalCondition_.concerningPlaces().size() == net.places_.size()) 
  {
    Condition tmpCond;
    tmpCond = net.finalCondition_.getFormula();
    set<const Place *> emptyPlaces = tmpCond.getFormula().getEmptyPlaces();
    
    if(emptyPlaces.size() == net.places_.size())
    {
      os << "ALL_PLACES_EMPTY;\n\n\n";
    }
    else
    {
      PNAPI_FOREACH(p, emptyPlaces)
      {
        tmpCond.removePlace(**p);
      }
      
      os << "(" << tmpCond << ") AND ALL_OTHER_PLACES_EMPTY;\n\n\n";
    }
  }
  else
  {
    os << net.finalCondition_ << ";\n\n\n";
  }

  return (os << delim("\n")
             << net.transitions_ << "\n\n{ END OF FILE '"
             << net.getMetaInformation(os, io::OUTPUTFILE) << "' }\n");
}


/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  os << p.getName();
  if((ModeData::data(os) == io::util::PLACE_TOKEN) && (p.getTokenCount() != 1))
  {
    os << ": " << p.getTokenCount();
  }
  
  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  os << "TRANSITION " << t.getName() << endl << delim(", ");

  if(t.getCost() != 0)
    os << "  COST " << t.getCost() << ";\n";

  if(!t.getRoles().empty() && !(util::RoleData::data(os).role))
  {
    os << "  ROLES\n    " << t.getRoles() << ";\n";
  }

  bool wroteInterfaceLabel = false;
  
  os << "  CONSUME\n    ";
  
  PNAPI_FOREACH(label, t.getLabels())
  {
    if(label->first->getType() == Label::INPUT)
    {
      os << (wroteInterfaceLabel ? ", " : "")
         << label->first->getName();
      
      if(label->second > 1)
      {
        os << ":" << label->second;
      }
  
      wroteInterfaceLabel = true;
    }
  }
  
  os << ((wroteInterfaceLabel && !t.getPresetArcs().empty()) ? ", " : "")
     << t.getPresetArcs() << ";\n"
     << "  PRODUCE\n    ";
  
  wroteInterfaceLabel = false;
  
  PNAPI_FOREACH(label, t.getLabels())
  {
    if(label->first->getType() == Label::OUTPUT)
    {
      os << (wroteInterfaceLabel ? ", " : "")
         << label->first->getName();
      
      if(label->second > 1)
      {
        os << ":" << label->second;
      }
  
      wroteInterfaceLabel = true;
    }
  }
  
  os << ((wroteInterfaceLabel && !t.getPostsetArcs().empty()) ? ", " : "")
     << t.getPostsetArcs() << ";\n";

  if(t.isSynchronized())
  {
    os << "  SYNCHRONIZE\n    " << t.getSynchronousLabels() << ";\n";
  }
  
  return os;
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  os << arc.getPlace().getName();
  if (arc.getWeight() > 1)
    os << ":" << arc.getWeight();
  return os;
}

/*!
 * \brief interface output
 */
std::ostream & output(std::ostream & os, const Interface & interface)
{
  os << "INTERFACE\n" << delim(", ");
  
  if(util::PortData::data(os).remove)
  {
    if(!interface.getInputLabels().empty())
    {
      os << "  INPUT\n    " << interface.getInputLabels() << ";\n"; 
    }
    if(!interface.getOutputLabels().empty())
    {
      os << "  OUTPUT\n    " << interface.getOutputLabels() << ";\n"; 
    }
    if(!interface.getSynchronousLabels().empty())
    {
      os << "  SYNCHRONOUS\n    " << interface.getSynchronousLabels() << ";\n"; 
    }
  }
  else
  {
    PNAPI_FOREACH(port, interface.getPorts())
    {
      os << (*port->second);  
    }
  }
  
  return (os << endl);
}

/*!
 * \brief port output
 */
std::ostream & output(std::ostream & os, const Port & port)
{
  os << "  PORT " << port.getName() << endl;
  
  if(!port.getInputLabels().empty())
  {
    os << "    INPUT\n      " << delim(", ") << port.getInputLabels() << ";\n"; 
  }
  if(!port.getOutputLabels().empty())
  {
    os << "    OUTPUT\n      " << port.getOutputLabels() << ";\n"; 
  }
  if(!port.getSynchronousLabels().empty())
  {
    os << "    SYNCHRONOUS\n      " << port.getSynchronousLabels() << ";\n"; 
  }
  
  return os;
}

/*!
 * \brief label output
 */
std::ostream & output(std::ostream & os, const Label & l)
{
  return (os << l.getName());
}

std::ostream & output(std::ostream & os,
    const std::pair<std::string, std::set<Place *> > & p)
    {
  return os << p.first << ": " << delim(", ") << p.second;
    }

/*!
 * \brief writing internal places grouped by capacity
 */
std::ostream & output(std::ostream & os,
                      const std::pair<unsigned int, std::set<Place *> > & p)
{
  if (p.first > 0)
  {
    os << "SAFE " << p.first << ": ";
  }
  
  return (os << delim(", ") << p.second << delim(";\n  "));
}

/*!
 * \brief negation output
 */
std::ostream & output(std::ostream & os, const formula::Negation & f)
{
  return (os << "NOT (" << **f.getChildren().begin() << ")");
}

/*!
 * \brief Conjunction output
 */
std::ostream & output(std::ostream & os, const formula::Conjunction & f)
{
  if (f.getChildren().empty())
    return (os << formula::FormulaTrue());
  else
    return (os << "(" << delim(" AND ") << f.getChildren() << ")");
}

/*!
 * \brief Disjunction output
 */
std::ostream & output(std::ostream & os, const formula::Disjunction & f)
{
  if (f.getChildren().empty())
    return (os << formula::FormulaFalse());
  else
    return (os << "(" << delim(" OR ") << f.getChildren() << ")");
}

/*!
 * \brief FormulaTrue output
 */
std::ostream & output(std::ostream & os, const formula::FormulaTrue &)
{
  return (os << "TRUE");
}

/*!
 * \brief FormulaFalse output
 */
std::ostream & output(std::ostream & os, const formula::FormulaFalse &)
{
  return (os << "FALSE");
}

/*!
 * \brief FormulaEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaEqual & f)
{
  return (os << f.getPlace().getName() << " = " << f.getTokens());
}

/*!
 * \brief FormulaNotEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaNotEqual & f)
{
  return (os << f.getPlace().getName() << " # " << f.getTokens());
}

/*!
 * \brief FormulaGreater output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreater & f)
{
  return (os << f.getPlace().getName() << " > " << f.getTokens());
}

/*!
 * \brief FormulaGreaterEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaGreaterEqual & f)
{
  return (os << f.getPlace().getName() << " >= " << f.getTokens());
}

/*!
 * \brief FormulaLess output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLess & f)
{
  return (os << f.getPlace().getName() << " < " << f.getTokens());
}

/*!
 * \brief FormulaLessEqual output
 */
std::ostream & output(std::ostream & os, const formula::FormulaLessEqual & f)
{
  return (os << f.getPlace().getName() << " <= " << f.getTokens());
}

} /* namespace __owfn */


/*************************************************************************
 ***** SA output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & sa(std::ios_base &base)
{
  util::FormatData::data(base) = util::SA;
  return base;
}


namespace __sa
{

/*!
 * \brief service automaton output
 */
std::ostream & output(std::ostream & os, const Automaton & sa)
{
  os << "INTERFACE" << endl << delim(", ")
     << "  INPUT\n    "
     << sa.getInputLabels()
     << ";\n"
     << "  OUTPUT\n    "
     << sa.getOutputLabels()
     << ";\n"
     << "  SYNCHRONOUS\n    "
     << sa.getSynchronousLabels()
     << ";\n\n"
     << "NODES\n"
     << delim("\n")
     << sa.states_;

  return (os << endl);
}

/*!
 * \brief state output
 */
std::ostream & output(std::ostream & os, const State & s)
{
  os << "  " << s.getName();
  if (s.isInitial() || s.isFinal())
  {
    os << ": "
       << (s.isInitial()
           ? (s.isFinal() ? "INITIAL, FINAL" : "INITIAL")
           : "FINAL");
  }

  return (os << endl << s.getPostsetEdges());
}

/*!
 * \brief edge output
 */
std::ostream & output(std::ostream & os, const Edge & e)
{
  return (os << "    " << e.getLabel() << " -> " << e.getDestination().getName());
}

} /* namespace __sa */

/*************************************************************************
 ***** Woflan output
 *************************************************************************/

/*!
 * \brief writes output type to stream 
 */
std::ios_base & woflan(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::WOFLAN;
  return ios;
}

namespace __woflan
{

/*!
 * \brief petri net output
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  string creator = net.getMetaInformation(os, CREATOR, PACKAGE_STRING);
  string inputfile = net.getMetaInformation(os, INPUTFILE);

  os //< output everything to this stream

  << "-- Petri net created by " << creator
  << (inputfile.empty() ? "" : " reading " + inputfile) << endl
  << endl

  << mode(io::util::PLACE) << endl
  << "  " << delim(";\n") << net.places_ << ";\n\n"

  // transitions
  << net.transitions_ << ";\n\n";

  return (os << "-- END OF FILE\n");
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  os << "place \"" << p.getName() << "\"";
  
  if (p.getTokenCount() > 0)
    os << "  init " << p.getTokenCount();
  
  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  return (os << "trans \"" << t.getName() << "\"\n"
            << delim(" ")
            << "  in "
            << t.getPresetArcs()
            << "\n  out "
            << t.getPostsetArcs() << endl);
}

/*!
 * \brief arc output
 */
std::ostream & output(std::ostream & os, const Arc & arc)
{
  for(unsigned int i = 0; i < arc.getWeight(); ++i)
  {
    os << "\"" << arc.getPlace().getName() << "\" ";
  }
  
  return os;
}

} /* namespace __woflan */


/*************************************************************************
 ***** INA output (*.pnt)
 *************************************************************************/

/*!
 * \brief writes output type to stream
 */
std::ios_base & pnt(std::ios_base & ios)
{
  util::FormatData::data(ios) = util::PNT;
  return ios;
}

namespace __pnt
{
/*!
 * \brief petri net output
 */
std::ostream & output(std::ostream & os, const PetriNet & net)
{
  os << "P   M   PRE,POST  NETZ 0\n"
     << util::mode(io::util::NORMAL) << net.places_
     << "@\n" << "place nr.             name capacity time\n"
     << util::mode(io::util::PLACE) << net.places_
     << "@\n" << "trans nr.             name priority time\n"
     << net.transitions_
     << "@\n";

  return os;
}

/*!
 * \brief place output
 */
std::ostream & output(std::ostream & os, const Place & p)
{
  if(ModeData::data(os) == io::util::NORMAL)
  {
    os << "  " << getNodeID(os, p) << " " << p.getTokenCount() << "    ";
    PNAPI_FOREACH(a, p.getPresetArcs())
    {
      os << getNodeID(os, (*a)->getSourceNode());
      if ((*a)->getWeight() > 1)
      {
        os << ": " << (*a)->getWeight();
      }
      os << " ";
    }
    if (p.getPostsetArcs().size() > 0)
    {
      PNAPI_FOREACH(a, p.getPostsetArcs())
      {
        os << getNodeID(os, (*a)->getTargetNode());
        if ((*a)->getWeight() > 1)
        {
          os << ": " << (*a)->getWeight();
        }
        os << " ";
      }
    }
    os << "\n";
  }
  else // ModeData::data(os) == io::util::PLACE
  {
    os << "       " << getNodeID(os, p) << ": " << getNodeName(os, p) << "      ";
    if (p.getCapacity() > 0)
    {
      os << " " << p.getCapacity();
    }
    else
    {
      os << "oo";
    }
    os << "   0\n";
  }

  return os;
}

/*!
 * \brief transition output
 */
std::ostream & output(std::ostream & os, const Transition & t)
{
  os << "       " << getNodeID(os, t) << ": " << getNodeName(os, t) << "      0   0\n";
  return os;
}

/*!
 * \brief get file internal node ID
 */
int getNodeID(std::ostream & os, const Node & n)
{
  Node * n_ = const_cast<Node *>(&n); // casting away the const
  if(util::PntData::data(os).ids.find(n_) == util::PntData::data(os).ids.end()) // if node not in cache
  {
    // get next free id and write to cache
    if (dynamic_cast<Place *>(n_) != NULL)
    {
      util::PntData::data(os).ids[n_] = util::PntData::data(os).placeID++;
    }
    else
    {
      util::PntData::data(os).ids[n_] = util::PntData::data(os).transitionID++;
    }
  }
  return util::PntData::data(os).ids[n_];
}

/*!
 * \brief get a short name of a node
 *
 * A node name must not be longer than 16 characters.
 * Truncating a name could lead to name collisions,
 * hence we have to find unused node names.
 *
 * \pre It is assumed that this method will only be called once for each node.
 */
std::string getNodeName(std::ostream & os, const Node & n)
{
  // get node id
  stringstream ss;
  ss << getNodeID(os, n);

  string result = ((dynamic_cast<const Place *>(&n) == NULL) ? "t" : "p") + ss.str();

  // fill up with whitespace
  while(result.length() < 16)
  {
    result = " " + result;
  }
  return result;
}

} /* namespace pnt */

} /* namespace io */

} /* namespace pnapi */
